;*****************************************************************************
;*** Name of the program:		TWI (Single) Master Implementation
;*** Function:             	Implementaion of I2C-Bus with connection to a
;***					   			1. I/O-Extension PCF8574
;***									2. DSM-0822a
;*** Author:	Andreas Hesse
;*** Date of construction:  	2003-04-09
;*** Date of last revision: 	2003-11-16
;*** Version:               	1.0.1.0
;*** Target:						AT90Sxxxx (any AVR device)
;*** Comment:
;***
;***	Abbreviations:
;***	1. RI = Remote Input
;***	2. RO = Remote Output
;***	3. PM = Program Memory
;***	4. SRAM = Static RAM
;******************************************************************************
;***
;*** Version management:
;*** V 1.0.1.0  Code Review 
;*** V 1.0.0.3  + I2C_Display 	 
;*** V 1.0.0.2  Optimized version: Read and write procedure are not separted
;***				 during write procedure for inputs also the output value will
;***				 be written
;*** V 1.0.0.1  Implement connection to PCF8574
;*** V 1.0.0.0  First construction based on  Version 1.0 of
;*** 		 A P P L I C A T I O N   N O T E   A V R 3 0 0 (2002.05.03)
;******************************************************************************
;*** DESCRIPTION
;*** 	Basic routines for communicating with TWI slave devices. This
;***	"single" master implementation is limited to one bus master on the
;***	TWI bus. Most applications do not need the multimaster ability
;***	the TWI bus provides. A single master implementation uses, by far,
;***	less resources and is less XTAL frequency dependent.
;***
;***	Some features :
;***	* All interrupts are free, and can be used for other activities.
;***	* Supports normal and fast mode.
;***	* Supports both 7-bit and 10-bit addressing.
;***	* Supports the entire AVR microcontroller family.
;***
;***	Main TWI functions :
;***	'TWI_start' -	Issues a start condition and sends address
;***						and transfer direction.
;***	'TWI_rep_start' -	Issues a repeated start condition and sends
;***							address and transfer direction.
;***	'TWI_do_transfer' -	Sends or receives data depending on
;***								direction given in address/dir byte.
;***	'TWI_stop' -	Terminates the data transfer by issue a
;***						stop condition.
;***
;*** USAGE
;***	Transfer formats is described in the AVR300 documentation.
;***	(An example is shown in the 'mainloop' code).
;***
;*** NOTES
;***	The TWI routines can be called either from non-interrupt or
;***	interrupt routines, not both.
;***
;*** STATISTICS
;***	Code Size: 		81 words (maximum)
;***	Register Usage: 	4 High, 0 Low
;***	Interrupt Usage: 	None
;***	Other Usage: 		Uses two I/O pins on port D
;***	XTAL Range: 		N/A
;***
;***************************************************************************

;*** Includes
;***  No Include

;**** Global I2C Constants ****

; Define I2C Ports and PINs
.equ	SCLP		= 4			; SCL Pin number on I2CPORT
.equ	SDAP		= 3			; SDA Pin number I2CPORT
.equ	I2CPIN 	= PINB		
.equ	I2CPORT 	= PORTB
.equ	I2CDDR 	= DDRB
	
.equ	b_dir	= 0			; transfer direction bit in tmpadr

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_hp_delay
;***	TWI_qp_delay
;***
;*** DESCRIPTION
;***	hp - half TWI clock period delay (normal: 5.0us / fast: 1.3us)
;***	qp - quarter TWI clock period delay (normal: 2.5us / fast: 0.6us)
;***
;***	SEE DOCUMENTATION !!!
;***
;*** USAGE
;***	no parameters
;***
;*** RETURN
;***	none
;***
;***************************************************************************

TWI_hp_delay:
	ldi	tmpdelay,9
TWI_hp_delay_loop:
	dec	tmpdelay
	brne	TWI_hp_delay_loop
	ret

TWI_qp_delay:
	ldi	tmpdelay,4
TWI_qp_delay_loop:
	dec	tmpdelay
	brne	TWI_qp_delay_loop
ret

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_rep_start
;***
;*** DESCRIPTION
;***	Assert repeated start condition and sends slave address.
;***
;*** USAGE
;***	tmpadr - Contains the slave address and transfer direction.
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to the address.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_start.
;***
;***************************************************************************

TWI_rep_start:
	sbi	I2CDDR,SCLP		; force SCL low
	cbi	I2CDDR,SDAP		; release SDA
	rcall	TWI_hp_delay	; half period delay
	cbi	I2CDDR,SCLP		; release SCL
	rcall	TWI_qp_delay	; quarter period delay


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_start
;***
;*** DESCRIPTION
;***	Generates start condition and sends slave address.
;***
;*** USAGE
;***	tmpadr - Contains the slave address and transfer direction.
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to the address.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_write.
;***
;***************************************************************************

TWI_start:
	mov	tmpdata,tmpadr	; copy address to transmitt register
	sbi	I2CDDR,SDAP		; force SDA low
	rcall	TWI_qp_delay	; quarter period delay


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_write
;***
;*** DESCRIPTION
;***	Writes data (one byte) to the TWI bus. Also used for sending
;***	the address.
;***
;*** USAGE
;***	tmpdata - Contains data to be transmitted.
;***
;*** RETURN
;***	Carry flag - Set if the slave respond transfer.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_get_ack.
;***
;***************************************************************************

TWI_write:
	sec						; set carry flag
	rol	tmpdata			; shift in carry and out bit one
	rjmp	TWI_write_first
TWI_write_bit:
	lsl	tmpdata			; if transmit register empty
TWI_write_first:
	breq	TWI_get_ack		; goto get acknowledge
	sbi	I2CDDR,SCLP		; force SCL low

	brcc	TWI_write_low	; if bit high
	nop						; (equalize number of cycles)
	cbi	I2CDDR,SDAP		; release SDA
	rjmp	TWI_write_high
TWI_write_low:				; else
	sbi	I2CDDR,SDAP		; force SDA low
	rjmp	TWI_write_high	; (equalize number of cycles)
TWI_write_high:
	rcall	TWI_hp_delay	; half period delay
	cbi	I2CDDR,SCLP		; release SCL
	rcall	TWI_hp_delay	; half period delay

	rjmp	TWI_write_bit

;***************************************************************************
;***
;*** FUNCTION
;***	TWI_get_ack
;***
;*** DESCRIPTION
;***	Get slave acknowledge response.
;***
;*** USAGE
;***	(used only by TWI_write in this version)
;***
;*** RETURN
;***	Carry flag - Cleared if a slave responds to a request.
;***
;***************************************************************************

TWI_get_ack:
	sbi	I2CDDR,SCLP		; force SCL low
	cbi	I2CDDR,SDAP		; release SDA
	rcall	TWI_hp_delay		; half period delay
	cbi	I2CDDR,SCLP		; release SCL

TWI_get_ack_wait:
	;sbis	I2CPIN,SCLP		; wait SCL high
	sbic	I2CPIN,SCLP		; wait SCL high
	rjmp 	TWI_ACK_OK		;(In case wait states are inserted)
	dec 	tmpcnt
	breq 	error
	rjmp	TWI_get_ack_wait
TWI_ACK_OK:
	clc						; clear carry flag
	sbic	I2CPIN,SDAP		; if SDA is high
	sec						; set carry flag
	rcall	TWI_hp_delay	; half period delay
TWI_ACK_END:
	ret

error:
	sbr	tmperror,0b00000001 	; Bit 1 is a Signal for an I2C-Bus error
	rcall ERRORHANDLING
	pop	tmp						; get return adress from stack
	rjmp  TWI_ACK_END


ERRORHANDLING:
	;cpi	tmperror,0
	;breq	I2CWRITE_END
	sbi	I2CPORT,3
	or		errorsave,tmperror
	clr	tmperror
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_do_transfer
;***
;*** DESCRIPTION
;***	Executes a transfer on bus. This is only a combination of TWI_read
;***	and TWI_write for convenience.
;***
;*** USAGE
;***	tmpadr - Must have the same direction as when TWI_start was called.
;***	see TWI_read and TWI_write for more information.
;***
;*** RETURN
;***	(depends on type of transfer, read or write)
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_read.
;***
;***************************************************************************

TWI_do_transfer:
	sbrs	tmpadr,b_dir	; if dir = write
	rjmp	TWI_write		;	goto write data


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_read
;***
;*** DESCRIPTION
;***	Reads data (one byte) from the TWI bus.
;***
;*** USAGE
;***	Carry flag - 	If set no acknowledge is given to the slave
;***			indicating last read operation before a STOP.
;***			If cleared acknowledge is given to the slave
;***			indicating more data.
;***
;*** RETURN
;***	tmpdata - Contains received data.
;***
;*** NOTE
;***	IMPORTANT! : This funtion must be directly followed by TWI_put_ack.
;***
;***************************************************************************

TWI_read:
	rol	tmpstatus		; store acknowledge
					; (used by TWI_put_ack)
	ldi	tmpdata,0x01	; data = 0x01
TWI_read_bit:			; do
	sbi	I2CDDR,SCLP	; force SCL low
	rcall	TWI_hp_delay	; half period delay

	cbi	I2CDDR,SCLP	; release SCL
	rcall TWI_hp_delay	; half period delay

	clc				; clear carry flag
	sbic	I2CPIN,SDAP	; if SDA is high
	sec				; set carry flag

	rol	tmpdata		; store data bit
	brcc	TWI_read_bit	; while receive register not full


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_put_ack
;***
;*** DESCRIPTION
;***	Put acknowledge.
;***
;*** USAGE
;***	(used only by TWI_read in this version)
;***
;*** RETURN
;***	none
;***
;***************************************************************************

TWI_put_ack:
	sbi	I2CDDR,SCLP			; force SCL low
	ror	tmpstatus				; get status bit
	brcc	TWI_put_ack_low	; if bit low goto assert low
	cbi	I2CDDR,SDAP			; release SDA
	rjmp	TWI_put_ack_high
TWI_put_ack_low:				; else
	sbi	I2CDDR,SDAP			; force SDA low
TWI_put_ack_high:

	rcall TWI_hp_delay		; half period delay
	cbi	 I2CDDR,SCLP			; release SCL
TWI_put_ack_wait:
	sbis	 I2CPIN,SCLP			; wait SCL high
	rjmp	 TWI_put_ack_wait
	rcall TWI_hp_delay		; half period delay
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_stop
;***
;*** DESCRIPTION
;***	Assert stop condition.
;***
;*** USAGE
;***	No parameters.
;***
;*** RETURN
;***	None.
;***
;***************************************************************************

TWI_stop:
	sbi	I2CDDR,SCLP		; force SCL low
	sbi	I2CDDR,SDAP		; force SDA low
	rcall	TWI_hp_delay		; half period delay
	cbi	I2CDDR,SCLP		; release SCL
	rcall	TWI_qp_delay		; quarter period delay
	cbi	I2CDDR,SDAP		; release SDA
	rcall	TWI_hp_delay		; half period delay
	ret


;***************************************************************************
;***
;*** FUNCTION
;***	TWI_init has been commented out; Initialisation will be performed in mainprogram
;***
;*** DESCRIPTION
;***	Initialization of the TWI bus interface.
;***
;*** USAGE
;***	Call this function once to initialize the TWI bus. No parameters
;***	are required.
;***
;*** RETURN
;***	None
;***
;*** NOTE
;***	I2CPORT and I2CDDR pins not used by the TWI bus interface will be
;***	set to Hi-Z (!).
;***
;*** COMMENT
;***	This function can be combined with other I2CPORT initializations.
;***
;***************************************************************************
;***     Portbit-Configuration	***

;***	I2CPORT bits
;***		1xxxxxx	   	PD6(Pin11)
;***		x1xxxxx	   	PD5(Pin09)
;***  	xx1xxxx	Set LED if signal from RI7 PD4(Pin08)
;***  	xxx1xxx	Error Connecting I2C  		PD3(Pin07)
;***  	xxxx1xx	Switch on Remote Output   	PD2(Pin06)
;***  	xxxxx1x 	I2C SCL  						PD1(Pin03)
;***  	xxxxxx1	I2C SDA							PD0(Pin02)
;***	PORTB bits
;***		1xxxxxxx		PB7(Pin19)
;***		x1xxxxxx		PB6(Pin18)
;***		xx1xxxxx		PB5(Pin17)
;***     xxx1xxxx	   PB4(Pin16)
;***     xxxx1xxx	   PB3(Pin15)
;***     xxxxx1xx	   PB2(Pin14)
;***     xxxxxx1x	   PB1(Pin13)
;***		xxxxxxx1   	PB0(Pin12)


;TWI_init:
;	ldi	tmpstatus,0b00000000	; PD0 and PD1 -> Input, No Pullup
;	out	I2CDDR,tmpstatus		; PD2 -> Input, switch on Pullup
;	ldi	tmpstatus,0b00000000	; PD3,PD4 -> Output
;	out	I2CPORT,tmpstatus		; set TWI pins to open colector
;	clr	tmpstatus				; clear TWI status register (used
									; as a temporary register)
;	ret



;***************************************************************************
;***
;*** FUNCTION
;***	I2CIO
;***
;*** DESCRIPTION
;***	Communcate with a PCF8574 Ciruit to expand I/O Lines
;***
;*** USAGE
;***	tmpadr - Adress of circuit
;***	are required.
;***
;*** RETURN
;***	None
;***
;***	I2CPORT and I2CDDR pins not used by the TWI bus interface will be
;*** NOTE
;***	set to Hi-Z (!).
;***
;*** COMMENT
;***	This function can be combined with other I2CPORT initializations.
;***
;***************************************************************************

;I2CIO:

;	ldi 	tmpcnt,$0F				; errorcounter, if no acknowledge
;	ldi	tmpadr,IOAdr+TWIwr	; Set device address and write
;	rcall TWI_start				; Send start condition and address
;	mov	tmpdata,tmp
;	rcall	TWI_do_transfer		; Execute transfer
;	ldi	tmpadr,IOAdr+TWIrd	; Set device address and read
;	rcall	TWI_rep_start			; Send repeated start condition and address
;	sec								; Set no acknowledge (read is followed by a
										; stop condition)
;	rcall	TWI_do_transfer		; Execute transfer (read)

;	rcall	TWI_stop					; Send stop condition - releases bus
;	ldi tmpcnt,0
;I2CEND:
;	ret
	
;***************************************************************************
;***
;*** FUNCTION
;***	I2CDISP
;***
;*** DESCRIPTION
;***	Communcate with a PCF8574 Ciruit to expand I/O Lines
;***
;*** USAGE
;***	tmpadr - Adress of circuit
;***	are required.
;***
;*** RETURN
;***	None
;***
;***	I2CPORT and I2CDDR pins not used by the TWI bus interface will be
;*** NOTE
;***	set to Hi-Z (!).
;***
;*** COMMENT
;***	This function can be combined with other I2CPORT initializations.
;***
;***************************************************************************

I2CDISP:					; gets the message in Z
	rcall	getmessage
I2CDISPWM:				; does not get message in Z automatically
							; must be done manually
	rcall GetCharDef
	ldi	yl,0x80
	ldi	yh,0
	ldi 	tmp16cntl,16
	ldi 	tmpcnt,0x80				; errorcounter, if no acknowledge
	ldi	tmpadr, I2CDispAdr + TWIwr
										; Set device address and write
	rcall TWI_start				; Send start condition and address
	ldi	tmpdata,224 			; select device 
	rcall	TWI_do_transfer		; Execute transfer
	ldi	tmpdata,200 			; device enable
	rcall	TWI_do_transfer		; Execute transfer
	ldi	tmpdata,240 			; Blink mode
	rcall	TWI_do_transfer		; Execute transfer
	ldi	tmpdata,0	 			; Initial-Position
	rcall	TWI_do_transfer		; Execute transfer
OutChar:	
	ld		tmpdata, y+		
	rcall	TWI_do_transfer		; 
	dec	tmp16cntl
	brne	OutChar
	mov	tmpdata,SpecialSig1	; last bytes represent
	rcall	TWI_do_transfer		; special signs on display
	mov	tmpdata,SpecialSig2	 			; 
	rcall	TWI_do_transfer		; 
	rcall	TWI_stop					; Send stop condition - releases bus

	ret
	
;***************************************************************************
;***
;*** FUNCTION
;*** Getmessage
;***
;*** DESCRIPTION
;***	Copies a message from PM to SRAM
;***
;*** USAGE
;*** Expects a pointer in Z to identify the Message and then loads 8 bytes from
;*** PM.
;***	xl,xh,yl,yh, tmp16cntl, tmp16cnth, tmpdelay
;***	are required.
;***
;*** RETURN
;***	None
;***
;*** NOTE
;***	
;***
;*** COMMENT
;***
;***
;***************************************************************************
	
GetMessage:
	ldi	xh,0
	ldi	xl,0x60		;StartSram
	ldi 	tmpdelay,8
	
CopyChar:
	lpm					; get value from PM
	st	 	X+,R0			; store it to SRAM
	adiw 	ZL,1 			; next location in PM
	dec	tmpdelay
	brne	CopyChar
ret

;***************************************************************************
;***
;*** FUNCTION
;*** GetCharDef
;***
;*** DESCRIPTION
;***	Copies the specific Byte codes for a Text from PM
;***  One Char on Display must be defined with 2 Bytes (2*8 Segments)
;***	Data Table CharDef holds the Segment definition
;*** USAGE
;*** Expects a pointer in Z to identify the Message and then loads 8 bytes from
;*** PM.
;***	xl,xh,yl,yh, tmp16cntl, tmp16cnth, tmpdelay
;***	are required.
;***
;*** RETURN
;***	None
;***
;*** NOTE
;***	
;***
;*** COMMENT
;***
;***
;***************************************************************************

GetCharDef:	
	ldi	xh,0
	ldi	xl,0x60		;StartSram
	ldi 	tmpdelay,8
	ldi	yl,0x80
	ldi	yh,0
CharDefLoop:
	; Load start of table
	ldi 	ZH,HIGH(2*CharDef) 
	ldi 	ZL,LOW(2*CharDef)
	; get position of  	
	ldi 	tmp16cnth,0
	ld		tmp16cntl, X+
	subi	tmp16cntl, 32	; convert ASCII to Position in Chardef-Table  
	lsl	tmp16cntl		; *2 (two bytes per char)	
	add	zl, tmp16cntl
	adc	zh, tmp16cnth
	lpm
	st		y+, r0
	adiw 	ZL,1 
	lpm
	st		y+, r0
	dec	tmpdelay
	brne 	CharDefLoop
	ret
	
;  Definitiontable
;	This Defines the values for the segments of DSM-0822a
; 	1 Char will need 2 Bytes (1 Char = 16 segements)
;  Additional X in comment indicates that a Char is worse to read on display

CharDef:

.db	0x00, 0x00 ;
.db	0x01, 0x20 ; ! XXX
.db	0x80, 0x20 ; "
.db 	0x25, 0x3d ; # XXX
.db	0xad, 0xb5 ; $ X
.db	0xec, 0x37 ; % X
.db	0x3b, 0x71 ; & XXX
.db	0x00, 0x02 ; '
.db	0x00, 0x42 ; (
.db	0x42, 0x00 ; )
.db	0x66, 0x66 ; *
.db	0x24, 0x24 ; +
.db	0x40, 0x00 ; ,
.db	0x20, 0x04 ; -
.db	0x01, 0x00 ; .
.db	0x40, 0x02 ; /
.db	0x99, 0x99 ; 0 gleic dem O
.db	0x00, 0x09 ; 1
.db	0x39, 0x9C ; 2
.db	0x29, 0x9D ; 3
.db   0xA0, 0x0D ; 4
.db	0xA9, 0x95 ; 5
.db	0xB9, 0x95 ; 6
.db	0x08, 0x89 ; 7
.db	0xB9, 0x9D ; 8
.db	0xA9, 0x9D ; 9
.db	0x28, 0x00 ; : XXX
.db	0x48, 0x00 ; ; XXX
.db	0x41, 0x12 ; < XX
.db	0x21, 0x14 ; = XXX
.db	0x03, 0x50 ; > XX
.db	0x8c, 0x8c ; ?
.db	0x3d, 0x99 ; @
.db	0xb8, 0x8d ; A
.db	0x0d, 0xbd ; B
.db	0x99, 0x90 ; C
.db	0x0d, 0xb9 ; D
.db	0xb9, 0x94 ; E
.db	0xb8, 0x80 ; F
.db	0x99, 0x95 ; G
.db	0xb0, 0x0d ; H
.db	0x0d, 0xb0 ; I
.db	0x1d, 0xa0 ; J
.db	0xb0, 0x42 ; K
.db	0x91, 0x10 ; L
.db	0x92, 0x0b ; M
.db	0x92, 0x49 ; N
.db	0x99, 0x99 ; O
.db	0xb8, 0x8c ; P
.db	0x99, 0xd9 ; Q
.db	0xb8, 0xcc ; R
.db	0xa9, 0x95 ; S
.db	0x0c, 0xa0 ; T
.db	0x91, 0x19 ; U
.db	0xd0, 0x02 ; V
.db	0xd0, 0x49 ; W
.db	0x42, 0x42 ; X
.db	0x06, 0x02 ; Y
.db	0x49, 0x92 ; Z
.db	0x99, 0x00 ; [
.db	0x02, 0x40 ; \ Backslas
.db	0x00, 0x99 ; ]
.db	0x40, 0x40 ; ^ XXX
.db	0x01, 0x10 ; _

;  Alternative fr kleine Zahlen, entsprecend einsetzen
;.db	0x9d, 0x20 ; 0
;.db	0x04, 0x20 ; 1
;.db	0x39, 0x20 ; 2
;.db	0x2d, 0x20 ; 3
;.db	0xa4, 0x20 ; 4
;.db	0xad, 0x00 ; 5
;.db	0xbd, 0x00 ; 6
;.db	0x0c, 0x20 ; 7
;.db	0xbd, 0x20 ; 8
;.db	0xad, 0x20 ; 9


;*** End of File